from cmath import inf
import neat
import pygame
import random
import os
import sys

WINDOW_WIDTH = 1000
WINDOW_HEIGHT = 600
GAME_HEIGHT = 500
OBSTACLE_SPACE_BETWEEN = 300

window_dim = [WINDOW_WIDTH, WINDOW_HEIGHT]

pygame.init()
window = pygame.display.set_mode(window_dim)
pygame.display.set_caption("Flappy bird")
font = pygame.font.SysFont('Comic Sans MS', 30)

class Obstacle():
    def __init__(self):
        self.width = 25
        self.height = random.randint(GAME_HEIGHT/10, GAME_HEIGHT/2)
        self.hole = random.randint(int(GAME_HEIGHT/6), int(GAME_HEIGHT/3))
        self.xpos = WINDOW_WIDTH
        self.ypos = GAME_HEIGHT - self.height - 20
        self.velocity = 2
        self.color = (255,255,255)
        
    def updatePosition(self):
        self.xpos -= self.velocity
        
    def drawObstacle(self):
        pygame.draw.rect(window, self.color, (self.xpos, self.ypos, self.width, self.height))
        pygame.draw.rect(window, self.color, (self.xpos, 0, self.width, GAME_HEIGHT - self.height - self.hole - 20))


class Floor():
    def __init__(self):
        self.height = 20
        self.image = pygame.transform.scale(pygame.image.load("floor.png"), (WINDOW_WIDTH, self.height))
        self.x = 0
        self.velocity = 2
    
    def draw(self):
        window.blit(self.image, (self.x, GAME_HEIGHT - self.height))
        window.blit(self.image, (WINDOW_WIDTH + self.x, GAME_HEIGHT - self.height))
        
    def update(self):
        self.x -= self.velocity
        if(abs(self.x) >= WINDOW_WIDTH):
            self.x = 0
    
class Ball():
    def __init__(self):
        self.xpos = 400
        self.ypos = GAME_HEIGHT / 2
        self.radius = 10
        self.color = (255,0,0)
        self.yvel = 0
        self.gravity = 0.3
        
    def draw(self):
        pygame.draw.circle(window, self.color, (self.xpos, self.ypos), self.radius)

    def update(self):       
        self.yvel += self.gravity
        
        self.ypos += self.yvel

        if(self.ypos < 0):
            self.ypos = 0
        if(self.ypos >= GAME_HEIGHT - self.radius - 20):
            self.ypos = GAME_HEIGHT - self.radius - 20

    
    def collide(self, obstacle):
        col_x = self.xpos + self.radius >= obstacle.xpos and obstacle.xpos + obstacle.width >= self.xpos - self.radius
        col_y = self.ypos + self.radius >= obstacle.ypos and obstacle.ypos + obstacle.height >= self.ypos - self.radius

        col_y2 = self.ypos + self.radius >= 0 and obstacle.ypos - obstacle.hole >= self.ypos - self.radius

        return col_x and (col_y or col_y2)
    
    def out_of_screen(self):
        if(self.ypos <= 0 or self.ypos >= GAME_HEIGHT - self.radius - 20):
            return True
        return False

    def jump(self):
        self.yvel = -7


def draw(obstacles, balls, floor, generation, fitness):
    window.fill((0,0,0))
    for o in obstacles:
        o.drawObstacle()
    for b in balls:
        b.draw() 
    floor.draw()

    text_gen = font.render("Generacija: " + str(generation), False, (255, 255, 255))
    
    text_fit = font.render("Fitness: " + str(fitness), False, (255, 255, 255))

    window.blit(text_gen, (0, 500))
    window.blit(text_fit, (0, 550))
    pygame.display.update()

generation = -1

def eval_genoms(genomes, config):
    global generation
    nns = []
    balls = []
    gen = []
    floor = Floor()

    generation += 1

    for id, genome in genomes :
        nns.append(neat.nn.FeedForwardNetwork.create(genome, config))
        balls.append(Ball())
        genome.fitness = 0
        gen.append(genome)

    obstacles = [Obstacle()]
    clock = pygame.time.Clock()
    ball_xpos = balls[0].xpos
    ball_radius = balls[0].radius

    while len(balls):
        clock.tick(200)
        for event in pygame.event.get():
            if event.type == pygame.QUIT:
                pygame.quit()
                sys.exit()
                    
        if(obstacles[len(obstacles) - 1].xpos <= WINDOW_WIDTH - OBSTACLE_SPACE_BETWEEN):
                obstacles.append(Obstacle())

        closest_obs = None
        minDistance = WINDOW_WIDTH

        for j, o in enumerate(obstacles):
            o.updatePosition()
            if (o.xpos + o.width <= 0):
                obstacles.pop(j)
                continue

            distance = o.xpos - ball_xpos
            
            if (distance >  -(ball_radius + o.width) and distance < minDistance):
                closest_obs = o
                minDistance = distance

        fitness = 0
        if(len(gen) > 0):
            fitness = round(gen[0].fitness, 2)

        for i, b in enumerate(balls):
            gen[i].fitness += 0.01
            
            if(b.collide(closest_obs) or b.out_of_screen()):
                    balls.pop(i)
                    nns.pop(i)
                    gen.pop(i)
                    continue

            dist_to_pipe = closest_obs.xpos - (b.xpos + b.radius)
            dist_to_upper_pipe = closest_obs.ypos - closest_obs.hole - (b.ypos - b.radius)
            dist_to_lower_pipe = closest_obs.ypos - (b.ypos + b.radius)

            nn = nns[i]

            #first experiment
            #output = nn.activate((b.ypos, b.yvel, dist_to_pipe, dist_to_lower_pipe, dist_to_upper_pipe))
            
            output = nn.activate((b.ypos / GAME_HEIGHT, dist_to_pipe / OBSTACLE_SPACE_BETWEEN, dist_to_lower_pipe / OBSTACLE_SPACE_BETWEEN, dist_to_upper_pipe / OBSTACLE_SPACE_BETWEEN))

            if(output[0] > 0.5):
                b.jump()

            b.update()
        
        floor.update()

        draw(obstacles, balls, floor, generation, fitness)

def run(config_file):
    config = neat.Config(neat.DefaultGenome, neat.DefaultReproduction, neat.DefaultSpeciesSet, neat.DefaultStagnation, config_file)
    p = neat.Population(config)
    p.add_reporter(neat.StdOutReporter(True))
    stats = neat.StatisticsReporter()
    p.add_reporter(stats)
    winner = p.run(eval_genoms, 30)
    stats.save_genome_fitness(',')
    stats.save_species_count(',')
    stats.save_species_fitness(',')
    with open('genom.txt', 'w') as file_gen:
        file_gen.write('{!s}'.format(winner))
        file_gen.write('')
        file_gen.write('{!s}'.format(stats.best_genome()))
    print('\nBest genome:\n{!s}'.format(winner))
    print('\nBest genome:\n{!s}'.format(stats.best_genome()))

if __name__ == "__main__":
    local_dir = os.path.dirname(__file__)
    config_path = os.path.join(local_dir, 'config-params.txt')
    run(config_path)